import math
import warnings

import pandas as pd
from dateutil.relativedelta import relativedelta
from datetime import datetime

from IRS_toolkit.core import cash_flow
from IRS_toolkit.utils import core,financial,schedule
from IRS_toolkit.utils.constants import VALID_CONVENTIONS
from IRS_toolkit.core.curve import yield_curve,compounded
warnings.filterwarnings("ignore")


class FloatLeg:
    """
    this class compute the float leg of the swap

        Args:
            nominal (float): notionel amount
            start_date (date): start date
            maturity_date (date): end date
            curve (curve): yield curve
            periodicity (int, optional): coupon payment periodicity. Defaults to 3.
            type_fill (str, optional): type of filling schedule. Defaults to "Forward".
    """

    def __init__(
        self,
        nominal,
        yield_curve_object:yield_curve.YieldCurve,
        schedule:schedule.Schedule=None,
        ESTR_coumpounded:compounded.Compounded=None,
        relative_delta=None,
    ):
        self.schedule = schedule
        self.date_format = schedule.date_format
        self.nominal = nominal
        self.start_date = schedule.start_date
        self.maturity_date = schedule.maturity_date
        self.yield_curve_object=yield_curve_object
        self.ESTR_coumpounded = ESTR_coumpounded
        self.cashflow = pd.DataFrame()
        
        self.accrued_coupon_float = 0
        self.date_convention = schedule.date_convention
        if relative_delta is None:
            relative_delta=relativedelta(days=0)
        self.relative_delta = relative_delta

    def compute_cash_flow(self, date_value:datetime):
        """
        this function computes the float cash flows

        Args:
            date_value (date): used to split the into accrued coupon and rest
        """

        self.cashflow = self.schedule.df
        self.cashflow["forward_zc"] = self.cashflow.apply(
            lambda x: self.yield_curve_object.forward_rates(
                x["start_date"], x["end_date"], self.relative_delta
            ),
            axis=1,
        )
        # here we want to know which coupon we are computing so that we can split to float coupon and accured coupon

        for i in range(0, len(self.cashflow.end_date), 1):
            if (
                pd.Timestamp(date_value) > list(self.cashflow.start_date)[i]
                and pd.Timestamp(date_value) < list(self.cashflow.end_date)[i]
            ):
                self.cashflow["forward_zc"][self.cashflow.index[i]] = (
                    self.yield_curve_object.forward_rates(
                        pd.Timestamp(date_value),
                        list(self.cashflow.end_date)[i],
                        self.relative_delta,
                    )
                )
                self.cashflow.day_count[self.cashflow.index[i]] = core.day_count(
                    pd.Timestamp(date_value),
                    list(self.cashflow.end_date)[i],
                    self.date_convention,
                )

        self.cashflow["forward_simple_rate"] = self.cashflow.apply(
            lambda x: financial.zc_to_simplerate(x["forward_zc"], x["day_count"]), axis=1
        )

        self.cashflow["cashflow"] = self.nominal * (
            (1 + self.cashflow.forward_zc) ** (self.cashflow.day_count) - 1
        )

        self.accrued_coupon_float = self.accrued_coupon(
            valuation_date=date_value
        )

    def discount_cashflow(
        self, discount_curve, date_valo=None, relative_delta=None
    ):
        """
        used to discount future cashflows to the date_valo

        Args:
            discount_curve (curve): yield curve
            date_valo (date, optional): valuation date. Defaults to None.
        """

        if relative_delta is None:
            relative_delta=relativedelta(days=0)

        dates = [date.strftime(self.date_format) for date in self.cashflow.end_date]

        amounts = self.cashflow.cashflow.to_list()

        discount_cashflow = cash_flow.cash_flow(dates, amounts)

        if date_valo:
            discount_cashflow.npv(date_valo, discount_curve, relative_delta)
        else:
            discount_cashflow.npv(discount_curve.date, discount_curve, relative_delta)

        self.cashflow[["discount_cashflow_amounts", "DF"]] = (
            discount_cashflow.cashflows[["discount_cashflow_amounts", "DF"]].values
        )

        self.NPV = discount_cashflow.NPV + self.accrued_coupon_float

    def accrued_coupon(self,
        valuation_date:datetime,
        relative_delta=None,
        ) -> float:
            """This function computes the accrued coupon of the float leg
                and for the past we use ESTR compounded and for the future we compute forwards

            Args:
                curve (curve): yield curve
                ESTR (dataframe): Estr compounded
                Cash_flows (Dataframe): dataframe
                notionel (float): float
                valuation_date (datetime): valuation date

            Returns:
                float: accrued coupon
            """

            if relative_delta is None:
                relative_delta=relativedelta(days=0)
            # if ESTR file is provided
            # we don't have weekends so we need to use interplation
            ESTR=self.ESTR_coumpounded.df
            date_min = min(ESTR["DATES"])
            date_max = max(ESTR["DATES"])
            SDate = financial.previous_coupon_date(self.cashflow["start_date"], self.cashflow["end_date"], pd.Timestamp(valuation_date))

            ESTR_start = ESTR[ESTR["DATES"] == SDate]["ESTR"]
            ESTR_end = ESTR[ESTR["DATES"] == valuation_date]["ESTR"]
            ESTR_max = ESTR[ESTR["DATES"] == date_max]["ESTR"]

            if (
                self.yield_curve_object.date_curve > SDate and date_max < SDate
                # Here my start Date is a Date in which no ESTR no FORWARD RATE (can't compute the forward)
            ):
                raise ValueError(
                    "Forward can't be computed (ex :Use an ESTR compounded up to curve date)"
                )

            result = 0


            if SDate < date_min or SDate > date_max:
                FRate = self.yield_curve_object.forward_rates(
                    financial.previous_coupon_date(self.cashflow["start_date"], self.cashflow["end_date"], pd.Timestamp(valuation_date)),
                    pd.Timestamp(valuation_date),
                    relative_delta,
                )

                Day_count_years = core.day_count(
                    financial.previous_coupon_date(self.cashflow["start_date"], self.cashflow["end_date"], pd.Timestamp(valuation_date)),
                    pd.Timestamp(valuation_date),
                    self.date_convention,
                )
                Perf = 0 if FRate is None else (1 + FRate) ** Day_count_years - 1
            elif valuation_date <= date_max:
                Perf = (float(ESTR_end) / float(ESTR_start)) - 1
            elif valuation_date > date_max:
                perf_0 = (float(ESTR_max) / float(ESTR_start)) - 1
                FRate0 = self.yield_curve_object.forward_rates(
                    pd.Timestamp(date_max) + timedelta(days=1),
                    pd.Timestamp(valuation_date),
                    relative_delta,
                )

                Day_count_years = core.day_count(
                    pd.Timestamp(date_max) + timedelta(days=1),
                    pd.Timestamp(valuation_date),
                    self.date_convention,
                )
                Perf = ((1 + FRate0) ** (Day_count_years) - 1) + perf_0 / self.nominal
            else:
                FRate = self.yield_curve_object.forward_rates(
                    financial.previous_coupon_date(self.cashflow["start_date"], self.cashflow["end_date"], pd.Timestamp(valuation_date)),
                    pd.Timestamp(valuation_date),
                    relative_delta,
                )

                Day_count_years = core.day_count(
                    financial.previous_coupon_date(self.cashflow["start_date"], self.cashflow["end_date"], pd.Timestamp(valuation_date)),
                    pd.Timestamp(valuation_date),
                )
                Perf = 0 if FRate is None else (1 + FRate) ** Day_count_years - 1
            result = self.nominal * Perf
            return result



# Standard library imports
import warnings
from datetime import datetime, timedelta

# Third-party imports
import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta

# Local/application imports
from IRS_toolkit.utils import core
from IRS_toolkit.utils.constants import VALID_COUPON_FREQUENCY,VALID_TENORS,VALID_CONVENTIONS,pay_frequency,TIMEZONE_PARIS

# Configure warnings
warnings.filterwarnings("ignore")


class YieldCurve:
    """
       A class for handling yield curves used in pricing and preparing
    zero-coupon curves for cash flow computation and discounting.
        Args:
               curve (dataframe): dataframe
               date_curve (date): date curve

        Attributs:
                date : curve date
                df : a dataframe that contains dates and ZC rates and Dfs

        Functions:
                setup() : create the dataframe and interpolate rate
                ForwardRate(begin,end) : compute the forward rate for two giving dates
                Bootstrap() : compute the ZC curve using the formula refer to notion appendix
                monthly_avg_daily() : compute the monthly average of instantaneous forward rates  using spot ZC curve


    """


    def __init__(self, list_tenor:list[str], list_rate:list[float], date_curve:datetime=None, convention:VALID_CONVENTIONS="ACT/360"):
        """Initialize a YieldCurve instance.
        
        Args:
            list_tenor (list[str]): List of tenors (e.g. ["1M", "3M", "6M", "1Y"])
            list_rate (list[float]): List of corresponding rates
            date_curve (datetime, optional): Reference date for the curve. Defaults to None.
            convention (VALID_CONVENTIONS, optional): Day count convention. Defaults to "ACT/360".
        """
        # Store input parameters
        self.list_tenor = list_tenor
        self.list_rate = list_rate
        self.date_curve = date_curve
        self.convention = convention

        # Create initial dataframe with input tenors and rates
        df = pd.DataFrame({
            "TENOR": ["0D"] + list_tenor,  # Add 0D tenor at start
            "STRIPPEDRATES": [np.nan] + list_rate  # Add NaN rate for 0D
        })

        # Add dates if date_curve is provided
        if date_curve is not None:
            df["AS_OF_DATE"] = date_curve
            df["DATE"] = pd.to_datetime(df["AS_OF_DATE"] + df["TENOR"].apply(core.tenor_to_period))
            
            # Calculate periods and day counts
            df["PERIOD"] = (df["DATE"] - self.date_curve).apply(lambda x: x.days)
            df["DAY_COUNT"] = df.apply(
                lambda x: core.day_count(self.date_curve, x["DATE"], convention=self.convention), 
                axis=1
            )

        df.sort_index(inplace=True)
        self.df = df

        # Create interpolated curve
        interpolated_df = df.copy()
        
        # Add relative date columns
        interpolated_df["RELATIVEDELTA"] = interpolated_df["TENOR"].apply(core.tenor_to_period)
        interpolated_df["RELATIVE_DATE"] = interpolated_df["RELATIVEDELTA"].apply(
            lambda x: (self.date_curve + x)
        )
        interpolated_df["PERIOD"] = interpolated_df["RELATIVE_DATE"].apply(
            lambda x: (x - self.date_curve).days
        )

        # Create daily periods and interpolate rates
        daily_periods = pd.DataFrame({
            "PERIOD": np.arange(1, max(interpolated_df["PERIOD"]) + 1)
        })
        interpolated_df = daily_periods.merge(interpolated_df, "left")
        interpolated_df["STRIPPEDRATES"] = interpolated_df["STRIPPEDRATES"].astype(float)
        interpolated_df["STRIPPEDRATES"].interpolate(method="cubic", inplace=True, limit_direction="forward")
        
        # Add dates and day counts
        interpolated_df["DATE"] = interpolated_df["PERIOD"].apply(lambda x: self.date_curve + timedelta(x))
        interpolated_df["DAY_COUNT"] = interpolated_df.apply(
            lambda x: core.day_count(self.date_curve, x["DATE"], convention=self.convention),
            axis=1
        )

        # Add 0D row at start
        zero_day = pd.DataFrame([{
            "PERIOD": 0,
            "TENOR": "0D",
            "STRIPPEDRATES": np.nan,
            "DATE": self.date_curve,
            "DAY_COUNT": 0.0
        }])
        
        interpolated_df = pd.concat([zero_day, interpolated_df], ignore_index=True)
        interpolated_df.sort_index(inplace=True)
        # Keep only needed columns in final interpolated curve
        self.df_interpolated = interpolated_df[["TENOR", "STRIPPEDRATES", "PERIOD", "DATE","DAY_COUNT"]]


    def forward_rates(
        self,
        begin:datetime,
        end:datetime,
        relative_delta=None,
        date_convention:VALID_CONVENTIONS="ACT/360",
        date_format="%Y-%m-%d",
    ):
        """
        compute forward rates

        Args:
            begin (date): start date
            end (date): end date

        Returns:
            float: forward rate
        """

        if relative_delta is None:
            relative_delta=relativedelta(days=0)

        # try:
        # Convert string dates to datetime, if necessary
        begin_date = datetime.strptime(begin, date_format) if isinstance(begin, str) else begin
        end_date = datetime.strptime(end, date_format) if isinstance(end, str) else end
        end_date = end_date + relative_delta

        # Validation of date ranges
        if end_date < self.date_curve or begin_date >= end_date:
            return None

        # Extract zero-coupon rates for the given dates
        zc_begin = self.df[self.df["DATE"] == begin_date.strftime("%Y-%m-%d")]["ZC"]
        zc_end = self.df[self.df["DATE"] == end_date.strftime("%Y-%m-%d")]["ZC"]

        if zc_begin.empty:
            return None  # Return None if no ZC rates found for the dates
        if zc_end.empty:
            return None  # Return None if no ZC rates found for the dates

        # Calculate discount factors (DF)
        num = (1 + zc_end.iloc[0]) ** core.day_count(self.date_curve, end_date,date_convention)
        den = (1 + zc_begin.iloc[0]) ** core.day_count(self.date_curve, begin_date,date_convention)
        result = (num / den) ** (
            1.0 / core.day_count(begin_date, end_date, date_convention)
        ) - 1

        # Compute forward rate using the formula (DF2/DF1)^(1/delta(t)) - 1
        return result

        # except Exception as e:
        #     print(f"An error occurred while calculating forward rates: {e}")
        #     return None  # Return a default value or None

    def bootstrap(
        self, coupon_frequency:VALID_COUPON_FREQUENCY, date_convention:VALID_CONVENTIONS, zc_curve=None
    ):
        """
        It Transform the yield curve to a zero-coupon (ZC) curve.

        This function processes the initial curve data to compute zero-coupon rates and discount factors.
        It handles different date calculations based on whether the current day is the first of the month.
        """
        if zc_curve is None:
            zc_curve = self.df.copy()

        coupon_periods = pay_frequency[coupon_frequency]*30
        coupon_frequency_date = pay_frequency[coupon_frequency]

        zc_date = [
            self.date_curve + relativedelta(months=i * coupon_frequency_date)
            for i in range(coupon_periods + 1)
        ]
        zc_curve_before = zc_curve[zc_curve["DATE"] < zc_date[1]]
        zc_curve_before["PERIOD"] = (zc_curve_before["DATE"] - self.date_curve).apply(
            lambda x: x.days
        )

        zc_curve_before["Coupon_period"] = zc_curve_before["DAY_COUNT"]

        zc_curve_before["ZC"] = (
            1 + zc_curve_before["STRIPPEDRATES"] * zc_curve_before["Coupon_period"]
        ) ** (1 / zc_curve_before["Coupon_period"]) - 1
        zc_curve_before["DF"] = (
            1 / (1 + zc_curve_before["ZC"]) ** (zc_curve_before["Coupon_period"])
        )

        zc_curve_temp = zc_curve[zc_curve["DATE"].isin(zc_date[1:])]
        zc_curve_temp.reset_index(drop=True, inplace=True)
        zc_curve_temp["Date_lagg"] = zc_curve_temp["DATE"].shift()
        zc_curve_temp["Date_lagg"].fillna(self.date_curve, inplace=True)

        zc_curve_temp["Coupon_period"] = zc_curve_temp.apply(
            lambda x: core.day_count(x["Date_lagg"], x["DATE"], date_convention), axis=1
        )
        zc_curve_temp["DF"] = 1
        for i in range(zc_curve_temp.shape[0]):
            zc_curve_temp.loc[i, "DF"] = (
                1
                - (
                    zc_curve_temp["STRIPPEDRATES"][i]
                    * zc_curve_temp["Coupon_period"]
                    * zc_curve_temp["DF"]
                )[:i].sum()
            ) / (
                1
                + zc_curve_temp["STRIPPEDRATES"][i] * zc_curve_temp["Coupon_period"][i]
            )
        zc_curve_temp["ZC"] = (1 / zc_curve_temp["DF"]) ** (
            1 / zc_curve_temp["DAY_COUNT"]
        ) - 1

        zc_curve = pd.concat([zc_curve_before, zc_curve_temp[zc_curve_before.columns]])
        zc_curve.reset_index(inplace=True, drop=True)
        self.df = zc_curve.merge(zc_curve.dropna(), "left")
        dates = pd.DataFrame(
            {
                "DATE": pd.date_range(
                    start=self.date_curve,
                    end=self.date_curve + relativedelta(years=30),
                    freq="D",
                ),
            }
        )

        self.df = dates.merge(zc_curve, "left")
        self.df["DF"] = self.df["DF"].astype(float)
        self.df["DF"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )
        self.df["PERIOD"] = (self.df["DATE"] - self.date_curve).apply(lambda x: x.days)
        self.df["DAY_COUNT"] = self.df.apply(
            lambda x: core.day_count(self.date_curve, x["DATE"], date_convention), axis=1
        )
        self.df["ZC"] = (1 / self.df["DF"]) ** (1 / self.df["DAY_COUNT"]) - 1
        self.df["STRIPPEDRATES"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )
        self.df.at[0, "DF"] = 1
        self.df.at[0, "Coupon_period"] = 0

    def monthly_avg_daily(
        self, start_date:datetime, end_date:datetime, frequency:str="D", relative_delta=None,date_format="%Y-%m-%d"
    ):
        """

        Args:
            start_date (date): start date
            end_date (date): end date

        Returns:
            Dataframe: Monthly average of daily forward rates
        """
        if relative_delta is None:
            relative_delta=relativedelta(days=0)

        if frequency == "Between Tenor":
            date_list = []
            for tenor in VALID_TENORS:
                date_forward = datetime.strptime(start_date, date_format) + core.tenor_to_period(tenor)
                date_list.append(date_forward)
        else:
            date_list = pd.date_range(start_date, end=end_date, freq=frequency)

        foreward_df = pd.DataFrame([date_list[:-1], date_list[1:]]).T
        foreward_df.columns = ["start_date", "end_date"]
        foreward_list = []
        for i, j in zip(date_list[:-1], date_list[1:]):
            foreward_list.append(self.ForwardRates(i, j, relative_delta))
        foreward_df["foreward_ZC"] = foreward_list
        foreward_df["day_count"] = foreward_df.apply(
            lambda x: core.day_count(x["start_date"], x["end_date"]), axis=1
        )
        foreward_df["foreward_simple"] = foreward_df.apply(
            lambda x: core.ZC_to_simplerate(x["foreward_ZC"], x["day_count"]), axis=1
        )

        foreward_df = foreward_df.set_index("start_date")
        foreward_df.index = pd.to_datetime(foreward_df.index)

        return foreward_df.groupby(pd.Grouper(freq="M")).mean(), foreward_df

    def bootstrap_12m_semi_yearly_coupon(self, coupon_frequency:VALID_COUPON_FREQUENCY):
        """
        Transform the yield curve to a zero-coupon (ZC) curve.

        This function processes the initial curve data to compute zero-coupon rates and discount factors.
        It handles different date calculations based on whether the current day is the first of the month.
        """
        zc_curve = self.df.copy()

        coupon_periods = {
            "quarterly": 29 * 4,
            "yearly": 29,
            "monthly": 29 * 12,
            "semi_annual": 29 * 2,
        }[coupon_frequency]
        coupon_frequency_date = {
            "quarterly": pd.DateOffset(months=3),  # "3MS",
            "yearly": pd.DateOffset(years=1),
            "monthly": pd.DateOffset(months=1),
            "semi_annual": pd.DateOffset(months=6),
        }[coupon_frequency]

        # if self.date.day == 1:
        zc_date1 = pd.date_range(
            self.date.strftime(self.date_format),
            periods=2,
            freq=pd.DateOffset(months=6),
        )

        zc_date2 = pd.date_range(
            (self.date + pd.DateOffset(years=1)).strftime(self.date_format),
            periods=coupon_periods,
            freq=coupon_frequency_date,
        )

        zc_date = zc_date1.append(zc_date2)

        zc_curve_before = zc_curve[zc_curve["Date"] < zc_date[1]]
        zc_curve_before["Period"] = (zc_curve_before["Date"] - self.date).apply(
            lambda x: x.days
        )

        zc_curve_before["Coupon_period"] = zc_curve_before["day_count"]

        zc_curve_before["ZC"] = (
            1 + zc_curve_before["StrippedRates"] * zc_curve_before["Coupon_period"]
        ) ** (1 / zc_curve_before["Coupon_period"]) - 1
        zc_curve_before["DF"] = (
            1 / (1 + zc_curve_before["ZC"]) ** (zc_curve_before["Coupon_period"])
        )

        zc_curve_temp = zc_curve[zc_curve["Date"].isin(zc_date)]
        zc_curve_temp.reset_index(drop=True, inplace=True)
        zc_curve_temp["Date_lagg"] = zc_curve_temp["Date"].shift()
        zc_curve_temp["Date_lagg"].fillna(self.date, inplace=True)
        zc_curve_temp["Coupon_period"] = zc_curve_temp.apply(
            lambda x: core.day_count(x["Date_lagg"], x["Date"]), axis=1
        )
        zc_curve_temp["DF"] = 1
        for i in range(zc_curve_temp.shape[0]):
            zc_curve_temp.loc[i, "DF"] = (
                1
                - (
                    zc_curve_temp["StrippedRates"][i]
                    * zc_curve_temp["Coupon_period"]
                    * zc_curve_temp["DF"]
                )[:i].sum()
            ) / (
                1
                + zc_curve_temp["StrippedRates"][i] * zc_curve_temp["Coupon_period"][i]
            )
        zc_curve_temp["ZC"] = (1 / zc_curve_temp["DF"]) ** (
            1 / zc_curve_temp["day_count"]
        ) - 1
        zc_curve = pd.concat([zc_curve_before, zc_curve_temp[zc_curve_before.columns]])
        zc_curve.reset_index(inplace=True, drop=True)
        self.df = self.df.merge(zc_curve.dropna(), "left")
        dates = pd.DataFrame(
            {
                "Date": pd.date_range(
                    start=self.date + relativedelta(days=1),
                    end=self.date + relativedelta(years=30),
                    freq="D",
                ),
            }
        )

        self.df = dates.merge(self.df, "left")
        self.df["DF"] = self.df["DF"].astype(float)
        self.df["DF"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )
        self.df["Period"] = (self.df["Date"] - self.date).apply(lambda x: x.days)
        self.df["day_count"] = self.df.apply(
            lambda x: core.day_count(self.date, x["Date"]), axis=1
        )
        self.df["ZC"] = (1 / self.df["DF"]) ** (1 / self.df["day_count"]) - 1
        self.df["StrippedRates"].interpolate(
            method="cubic", inplace=True, limit_direction="forward"
        )



from datetime import datetime
from dateutil.relativedelta import relativedelta
from IRS_toolkit.utils.constants import VALID_FILL_TYPE,VALID_CONVENTIONS
from IRS_toolkit.utils import core
import pandas as pd
import math

class Schedule:
    def __init__(self,list_date:list[list]=None,start_date:datetime=None,maturity_date:datetime=None,periodicity:int=None,type_fill:VALID_FILL_TYPE="Forward",date_convention:VALID_CONVENTIONS="ACT/360",date_format="%Y-%m-%d"):

        if list_date is not None and len(list_date)>0 and len(list_date[0])>1:
            self.list_date=list_date
            if start_date is None:
                start_date=list_date[0][0]
            if maturity_date is None:
                maturity_date=list_date[0][-1]
            self.df=pd.DataFrame()
            for i in list_date.copy():
                self.add_date_range(i)
        else:

            self.list_date=[]
            def convert(strg):
                if type(strg) is str:
                    if strg == "NaT":
                        return "N/A"
                    return datetime.strptime(strg, self.date_format)
                else:
                    return strg

            delta = relativedelta(convert(maturity_date), convert(start_date))
            months = (delta.years * 12 + delta.months + delta.days / 30) / periodicity
            period = math.ceil(months)

            if type_fill == "Forward":
                start_range = pd.date_range(
                    start=start_date, periods=period, freq=pd.DateOffset(months=periodicity)
                )

                end_range = start_range + pd.DateOffset(months=periodicity)

                self.df = pd.DataFrame(
                    {"start_date": start_range, "end_date": end_range}
                )

                for i in range(1, len(self.df), 1):
                    if self.df["start_date"][i] != self.df["end_date"][i - 1]:
                        self.df["start_date"][i] = self.df["end_date"][i - 1]

            elif type_fill == "Back":
                end_range = pd.date_range(
                    end=maturity_date, periods=period, freq=pd.DateOffset(months=periodicity)
                )

                start_range = end_range - pd.DateOffset(month=periodicity)

                self.df = pd.DataFrame(
                    {"start_date": start_range, "end_date": end_range}
                )

                for i in range(1, len(self.df), 1):
                    if self.df["start_date"][i] != self.df["end_date"][i - 1]:
                        self.df["start_date"][i] = self.df["end_date"][i - 1]
            else:
                raise ValueError("Error: Not an available option.")

            for _index, row in self.df.iterrows():
                self.add_date_range([row['start_date'],row['end_date']])

            self.df["start_date"].iloc[0] = pd.Timestamp(start_date)
            self.df["end_date"].iloc[-1] = pd.Timestamp(maturity_date)
        

        self.df["Period"] = (
            self.df.end_date - self.df.start_date
        ).apply(lambda x: x.days)

        self.df["day_count"] = self.df.apply(
            lambda x: core.day_count(
                x["start_date"], x["end_date"], date_convention
            ),
            axis=1,
        )

        self.date_format=date_format
        self.start_date=start_date
        self.maturity_date=maturity_date
        self.periodicity=periodicity
        self.type_fill=type_fill
        self.date_convention=date_convention
    
    def add_date_range(self,date_range:list[datetime]):
        self.list_date.append(date_range)
        sub_df=pd.DataFrame.from_dict(data={"start_date":[date_range[0]],"end_date":[date_range[1]]})
        self.df=pd.concat([self.df,sub_df])



# Standard library imports
import calendar
from datetime import date, datetime, timedelta
from typing import Union

# Third-party library imports
import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from IRS_toolkit.utils.constants import VALID_CONVENTIONS, TIMEZONE_PARIS
# Constants

def day_count(start:Union[date, datetime], end:Union[date, datetime], convention:VALID_CONVENTIONS) -> float:
    """
    This fucntion computes the period in years between two given dates
            with a defined convention
    Args:
        start (datetime): start date
        end (datetime): end date
        convention (str, optional): day count convention. Defaults to "ACT/360".

    Returns:
        float: day count with the given  day count convention
    """

    result=0

    if end < start:
        raise ValueError("End date must be after start date")

    if end == start:
        result = 0.0
    if convention == "ACT/360":
        result = (end - start).days / 360

    elif convention == "ACT/ACT":
        start_dt=start
        end_dt=end
        if start_dt.year == end_dt.year:
            days_in_year = 366 if calendar.isleap(start_dt.year) else 365
            days = (end_dt - start_dt).days
            result= days / days_in_year
        else:
            # Calculate for different years
            result = 0.0

            # First partial year
            year1_end = datetime(start_dt.year + 1, 1, 1)
            days_year1 = 366 if calendar.isleap(start_dt.year) else 365
            result += (year1_end - start_dt).days / days_year1

            # Full years in between
            result += end_dt.year - start_dt.year - 1

            # Last partial year
            year2_start = datetime(end_dt.year, 1, 1)
            days_year2 = 366 if calendar.isleap(end_dt.year) else 365
            result += (end_dt - year2_start).days / days_year2


    elif convention == "30/360":
        # Ensure start_date is before end_date
        if start > end:
            start, end = end, start

        # Extract year, month, day from the dates
        start_year, start_month, start_day = start.year, start.month, start.day
        end_year, end_month, end_day = end.year, end.month, end.day

        # Adjust days for 30/360 calculation
        if start_day == 31 or (
            start_month == 2 and start_day in (29,28)
        ):
            start_day = 30
        if end_day == 31 and start_day == 30:
            end_day = 30

        # Calculate the difference in days
        result = (
            (end_year - start_year) * 360
            + (end_month - start_month) * 30
            + (end_day - start_day)
        ) / 360
    return result

def linear_interpolation(list_dates:list[datetime], list_values:list[float]) -> pd.DataFrame:
    """this function return a dataframe filled with the missing dates and
    calculate the linear interpolation on the values column.

    Args:
        df (DataFrame): the original DFs
        date_column (str, optional): Date. Defaults to 'date'.
        value_column (str, optional): value. Defaults to 'ESTR'.

    Returns:
        _type_: _description_
    """

    complete_dates = pd.date_range(
        start=min(list_dates,default=None), end=max(list_dates,default=None), freq="D"
    )

    def to_timestamp(dt):
        """Convert date or datetime to timestamp."""
        if isinstance(dt, datetime):
            return dt.timestamp()
        elif isinstance(dt, date):
            # Convert date to datetime at midnight
            return datetime.combine(dt, datetime.min.time()).timestamp()
        else:
            print(type(dt))
            raise TypeError("Input must be a date or datetime object.")

    # steps
    date_column_numerical = [to_timestamp(d) for d in list_dates]
    complete_dates_numerical = [to_timestamp(d) for d in complete_dates]

    interpolated_values = np.interp(
        complete_dates_numerical, date_column_numerical, list_values
    )
    dates = [dt.strftime("%Y-%m-%d") for dt in complete_dates]
    interpolated_dict={dates[i]: interpolated_values[i] for i in range(len(dates))}
    interpolated_df=pd.DataFrame.from_dict(interpolated_dict,orient="index",columns=["VALUES"])
    interpolated_df["DATES"]=interpolated_df.index

    return interpolated_df

def tenor_to_period(tenor: str) -> Union[timedelta, relativedelta]:
    """
    Convert a given tenor to a period.

    Args:
        tenor (str): A string representing the tenor (e.g., '1D', '2W', '3M', '1Y').

    Returns:
        Union[timedelta, relativedelta]: The corresponding period as a timedelta or relativedelta object.

    Raises:
        ValueError: If the tenor unit is invalid.

    Example:
        >>> tenor_to_period('1D')
        datetime.timedelta(days=1)
        >>> tenor_to_period('2W')
        datetime.timedelta(days=14)
        >>> tenor_to_period('3M')
        relativedelta(months=+3)
    """
    # Extract numeric value and unit from the tenor
    tenor_value = int(tenor[:-1])
    tenor_unit = tenor[-1].lower()

    # Define a dictionary mapping tenor units to their corresponding period objects
    dict_tenor = {
        'd': timedelta(days=tenor_value),
        'w': timedelta(weeks=tenor_value),
        'm': relativedelta(months=tenor_value),
        'y': relativedelta(years=tenor_value)
    }

    # Return the corresponding period if the unit is valid, otherwise raise an error
    if tenor_unit in dict_tenor:
        return dict_tenor[tenor_unit]
    else:
        raise ValueError(f"Invalid tenor unit: {tenor_unit}. Valid units are 'd', 'w', 'm', 'y'.")


def period_to_tenor(period: int) -> str:
    """
    Convert a given period in days to its corresponding tenor.

    Args:
        period (int): Number of days.

    Returns:
        str: Corresponding tenor, or None if no match is found.

    Note:
        This function assumes 30 days per month and 360 days per year.
    """
    # Ensure period is an integer
    period = int(period)

    # Define tenor dictionary with optimized calculations
    tenor_dict = {
        1: "1D", 7: "1W", 14: "2W", 21: "3W",
        **{30 * i: f"{i}M" for i in range(1, 12)},  # 1M to 11M
        360: "1Y",
        360+90: "15M", 360+180: "18M", 360+270: "21M",
        **{360 * i: f"{i}Y" for i in range(2, 13)},  # 2Y to 12Y
        360 * 15: "15Y", 360 * 20: "20Y", 360 * 25: "25Y", 360 * 30: "30Y"
    }
    if period in tenor_dict:
        result=tenor_dict.get(period)
    else:
        raise ValueError(f"Invalid period: {period}. Valid periods are {list(tenor_dict.keys())}.")

    # Return the tenor if found, otherwise None
    return result

def previous_coupon_date(list_start_dates:list[datetime], list_end_dates:list[datetime], valuation_date:datetime) -> datetime:
    """
    get the pervious coupon date with a given valuation date

    Args:
        df (Dataframe): Payments  details
        valuation_date (datetime): valuation date

    Returns:
        datetime: previous coupon date
    """
    list_start_dates=list(list_start_dates)
    list_end_dates=list(list_end_dates)
    previous_coupon = valuation_date

    for date_index in range(len(list_start_dates)):
        if valuation_date >= list_start_dates[date_index] and valuation_date < list_end_dates[date_index]:
            previous_coupon=list_start_dates[date_index]
    return previous_coupon


from typing import Literal
from dateutil.relativedelta import relativedelta
from datetime import date, datetime, timedelta
from zoneinfo import ZoneInfo

VALID_CONVENTIONS = Literal["ACT/360", "ACT/ACT","30/360"]
pay_frequency={"monthly":1, "quarterly":3, "semi_annual":6, "annual":12}
VALID_TENORS = Literal["D","W","M","Y"]
VALID_COUPON_FREQUENCY = Literal["quarterly","annual", "semi_annual", "monthly"]
VALID_TENORS = Literal[
            "1D",
            "1W",
            "2W",
            "3W",
            "1M",
            "2M",
            "3M",
            "4M",
            "5M",
            "6M",
            "7M",
            "8M",
            "9M",
            "10M",
            "11M",
            "1Y",
            "15M",
            "18M",
            "21M",
            "2Y",
            "3Y",
            "4Y",
            "5Y",
            "6Y",
            "7Y",
            "8Y",
            "9Y",
            "10Y",
            "11Y",
            "12Y",
            "15Y",
            "20Y",
            "25Y",
            "30Y",
]
VALID_FILL_TYPE = Literal["Back", "Forward"]

TIMEZONE_PARIS = ZoneInfo("Europe/Paris")


import math
import warnings
from datetime import datetime

import pandas as pd
from dateutil.relativedelta import relativedelta
from datetime import datetime

from IRS_toolkit.core import cash_flow
from IRS_toolkit.utils import core,financial
from IRS_toolkit.utils.constants import VALID_FILL_TYPE,VALID_CONVENTIONS
from IRS_toolkit.utils import schedule
from IRS_toolkit.core.curve import yield_curve
warnings.filterwarnings("ignore")


class FixLeg:
    """
    this class compute the float leg of the swap

        Args:
            nominal (float): notionel amount
            start_date (date): start date
            maturity_date (date): end date
            fix_rate (float): swap fixed rate
            periodicity (int, optional): coupon payment periodicity. Defaults to 3.
            type_fill (str, optional): type of filling schedule. Defaults to "Forward".
    """

    def __init__(
        self,
        nominal:float,
        fix_rate:float,
        schedule:schedule.Schedule,
    ):
        self.date_format = schedule.date_format
        self.nominal = nominal
        self.fix_rate = fix_rate
        self.cashflow = pd.DataFrame()
        self.accrued_coupon_fix = 0
        self.spread = 0
        self.date_convention = schedule.date_convention
        self.schedule = schedule

    def compute_cash_flow(self, date_value, spreadHC=0.0, spreadGC=0.0):
        """
        this function compute the float cash flows

        """
        self.cashflow = self.schedule.df
        self.cashflow["cashflow"] = (
            float(self.nominal) * float(self.fix_rate) * self.cashflow["day_count"]
        )
        """        for i in range(0, len(self.cashflow.end_date) - 1, 1):
            if (
                pd.Timestamp(date_value) > self.cashflow.start_date[i]
                and pd.Timestamp(date_value) < self.cashflow.end_date[i]
            ):
                self.cashflow.day_count[i] = day_count(
                    pd.Timestamp(date_value), self.cashflow.end_date[i]
                )
                self.cashflow["cashflow"][i] = (
                    self.nominal * self.fix_rate * self.cashflow.day_count[i]
                )"""

        coupon_period = core.day_count(
            core.previous_coupon_date(self.cashflow["start_date"], self.cashflow["end_date"], pd.Timestamp(date_value)),
            pd.Timestamp(date_value),
            self.date_convention,
        )
        self.accrued_coupon_fix = self.nominal * self.fix_rate * coupon_period
        self.spreadHC = financial.spread_amount(
            self.cashflow["start_date"], self.cashflow["end_date"], self.nominal, spreadHC, valuation_date=date_value, convention=self.date_convention
        )
        self.spreadGC = financial.spread_amount(
            self.cashflow["start_date"], self.cashflow["end_date"], self.nominal, spreadGC, valuation_date=date_value, convention=self.date_convention
        )
        self.spread = self.spreadGC + self.spreadHC

    def discount_cashflow(
        self, discount_curve:yield_curve.YieldCurve, date_valo=None, relative_delta=None
    ):
        """
        used to discount future cashflows to the date_valo

        Args:
            discount_curve (curve): yield curve
            date_valo (date, optional): valuation date. Defaults to None.
        """
        relative_delta=relativedelta(days=0)

        dates = [date_obj.strftime(self.date_format) for date_obj in self.cashflow.end_date]

        amounts = self.cashflow.cashflow.to_list()

        discount_cashflow = cash_flow.cash_flow(dates, amounts)

        if date_valo:
            discount_cashflow.npv(date_valo, discount_curve, relative_delta)
        else:
            discount_cashflow.npv(discount_curve.date_curve, discount_curve, relative_delta)
        self.cashflow[["discount_cashflow_amounts", "DF"]] = (
            discount_cashflow.cashflows[["discount_cashflow_amounts", "DF"]].values
        )
        self.NPV = discount_cashflow.NPV



import warnings

import pandas as pd
from scipy.optimize import minimize

from IRS_toolkit.core.leg import fix_leg, float_leg
from datetime import datetime
from dateutil.relativedelta import relativedelta
from IRS_toolkit.core.curve import compounded, yield_curve
from IRS_toolkit.utils.constants import VALID_FILL_TYPE,VALID_CONVENTIONS
from IRS_toolkit.utils import schedule

warnings.filterwarnings("ignore")


class Swap:
    """
    A class that provides various outputs related to swap pricing.


    Args:
        fix_leg (legFix): fixed leg
        float_leg (legFloat): float leg
    """

    def __init__(self,
                nominal:float,
                fix_rate:float=None,
                yield_curve:yield_curve.YieldCurve=None,
                ESTR_compounded:compounded.Compounded=None,
                schedule_fix:schedule.Schedule=None,
                schedule_float:schedule.Schedule=None,
                relative_delta=None):

        self.nominal=nominal
        self.fix_rate=fix_rate
        self.yield_curve=yield_curve
        self.ESTR_compounded=ESTR_compounded
        self.schedule_fix=schedule_fix
        self.schedule_float=schedule_float
        self.relative_delta=relative_delta

        fix_leg_object=fix_leg.FixLeg(nominal=nominal,
        fix_rate=fix_rate,
        schedule=schedule_fix)

        float_leg_object=float_leg.FloatLeg(nominal=nominal,
                                            yield_curve_object=yield_curve,
                                            schedule=schedule_float,
                                            ESTR_coumpounded=ESTR_compounded,
                                            relative_delta=relative_delta)

        self.fix_leg_object = fix_leg_object
        self.float_leg_object = float_leg_object

    def npv(self, valuation_date:datetime):
        """
        Net present value of the swap

        Args:
            discount_curve (curve): yield curve
            date_valo (date, optional): valuation date. Defaults to None.
            RunningCouv (float, optional): spread. Defaults to 0.00.
            GainGC (float, optional): spread. Defaults to 0.00.

        Returns:
            float: Net present value
        """
        self.fix_leg_object.discount_cashflow(self.yield_curve, valuation_date)
        self.float_leg_object.discount_cashflow(self.yield_curve, valuation_date)
        self.NPV_ = self.fix_leg_object.NPV - self.float_leg_object.NPV
        return self.NPV_

    def fair_rate(self, valuation_date:datetime):
        """
        fair rate of the swap

        Args:
            date_valo (date): date valuation
            ImpSchedule (dataframe) : in case you use imported schedule

        Returns:
            float, float: fair rate, theorical fair rate
        """
        fix_rate = self.fix_leg_object.fix_rate

        fix_leg_object_nominal=self.fix_leg_object.nominal
        fix_leg_object_schedule=self.fix_leg_object.schedule

        def loss_func(fix_rate:float):
            leg_fix = fix_leg.FixLeg(
                nominal=fix_leg_object_nominal,
                fix_rate=fix_rate,
                schedule=fix_leg_object_schedule,
            )
            leg_fix.compute_cash_flow(pd.Timestamp(valuation_date))
            leg_fix.discount_cashflow(self.float_leg_object.yield_curve_object, valuation_date)
            return (leg_fix.NPV - self.float_leg_object.NPV) * (
                leg_fix.NPV - self.float_leg_object.NPV
            )

        res = minimize(
            loss_func,
            fix_rate,
            method="nelder-mead",
            options={"xatol": 1e-8, "disp": True},
        )
        self.faire_rate = float(res.x)
        self.faire_rate_theory = (
            self.float_leg_object.NPV
            / (
                self.fix_leg_object.nominal
                * self.fix_leg_object.cashflow.iloc[:, 3]
                * self.fix_leg_object.cashflow.DF
            ).sum()
        )
        return self.faire_rate, self.faire_rate_theory

    def price(self, valuation_date:datetime,spreadHC:float,spreadGC:float):
        if self.relative_delta is None:
            relative_delta=relativedelta(days=0)
        else:
            relative_delta=self.relative_delta

        self.fix_leg_object.compute_cash_flow(valuation_date, spreadHC, spreadGC)
        self.fix_leg_object.discount_cashflow(self.yield_curve, valuation_date, relative_delta)

        self.float_leg_object.compute_cash_flow(valuation_date)
        self.float_leg_object.discount_cashflow(self.yield_curve, valuation_date, relative_delta)



import warnings

import pandas as pd
from dateutil.relativedelta import relativedelta
from datetime import datetime, timedelta

from IRS_toolkit.utils import core
from IRS_toolkit.core.curve import yield_curve
from IRS_toolkit.utils.constants import VALID_CONVENTIONS

warnings.filterwarnings("ignore")


class cash_flow:
    """
    cash flows handling
        Args:
            dates (Datframe): dates
            amounts (Dataframe): coupons
    """

    def __init__(self, dates:list[datetime], amounts:list[float],date_format="%Y-%m-%d",date_convention:VALID_CONVENTIONS="ACT/360"):
        dates = [datetime.strptime(dt, date_format) if isinstance(dt, str) else dt for dt in dates]
        self.cashflows = pd.DataFrame(
            {"cash_flow_date": dates, "cash_flow_amount": amounts}
        )
        self.date_convention = date_convention
    def npv(self, valuation_date:datetime, curve:yield_curve.YieldCurve, relative_delta=None,date_format="%Y-%m-%d"):
        """
        Compute the Net present value

        Args:
            valuation_date (date): valuation date
            curve (curve): yield curve

        Returns:
            float: Net present value of future cash flows
        """
        if relative_delta is None:
            relative_delta=relativedelta(days=0)
        self.cashflows["discount_cashflow_amounts"] = 0
        self.cashflows["DF"] = 0

        valuation_date = datetime.strptime(valuation_date, date_format) if isinstance(valuation_date, str) else valuation_date
        valuation_date = valuation_date + timedelta(days=1)

        for ind, dt in self.cashflows.iterrows():
            if dt.cash_flow_date > valuation_date:
                self.cashflows.loc[ind, "DF"] = 1 / (
                    1
                    + curve.forward_rates(
                        valuation_date, dt.cash_flow_date, relative_delta
                    )
                ) ** (core.day_count(valuation_date, dt.cash_flow_date,self.date_convention))

        self.cashflows["discount_cashflow_amounts"] = (
            self.cashflows.DF * self.cashflows.cash_flow_amount
        )
        self.NPV = self.cashflows["discount_cashflow_amounts"].sum()
        return self.cashflows["discount_cashflow_amounts"].sum()

    # wighted present value for bonds
    def wpv(self, valuation_date:datetime, curve:yield_curve.YieldCurve, relative_delta=None,date_format="%Y-%m-%d"):
        """
        Weighted Net present value

        Args:
            valuation_date (date): valuation date
            curve (curve): yield curve

        Returns:
            float: time weightide Net present value of future cash flows
        """
        if relative_delta is None:
            relative_delta=relativedelta(days=0)

        self.cashflows["wdiscount_cashflow_amounts"] = 0
        self.cashflows["DF"] = 0

        valuation_date = datetime.strptime(valuation_date, date_format) if isinstance(valuation_date, str) else valuation_date

        for ind, dt in self.cashflows.iterrows():
            if dt.cash_flow_date > valuation_date:
                self.cashflows.loc[ind, "DF"] = 1 / (
                    1
                    + curve.forward_rates(
                        valuation_date, dt.cash_flow_date, relative_delta
                    )
                ) ** (core.day_count(valuation_date, dt.cash_flow_date))

        self.cashflows["wdiscount_cashflow_amounts"] = (
            self.cashflows.DF
            * self.cashflows.cash_flow_amount
            * self.cashflows.cash_flow_date.apply(
                lambda x: core.day_count(valuation_date, x, self.date_convention)
            )
        )
        self.WPV = (self.cashflows["wdiscount_cashflow_amounts"]).sum()
        return self.cashflows["wdiscount_cashflow_amounts"].sum()
